import yaml
import importlib

import collections
import functools
from schema import Schema, Optional, Or, And
from .context import Context

# From https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize


class InvalidConfig(Exception):
    ...


_supported_manifest = ["mpd"]


class Config(object):

    _schema = Schema(
        [  # Must start with a list of valid formats
            {
                Or(*_supported_manifest):
                [
                    {
                        # plugins path: configuration
                        str: Or(list, dict)
                    }
                ]
            }
        ]
    )

    @ staticmethod
    def validate(config):
        if not Config._schema.is_valid(config):
            raise InvalidConfig('Invalid .yml pipeline configuration')


class Memoized(object):
    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not isinstance(args, collections.Hashable):
            # uncacheable. a list, for instance.
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)


class Pipeline:

    # TODO needs to be shared and thus elsewhere. Notice that support for
    # m3u8 is missing at the moment
    SUPPORTED_MANIFEST_FORMATS = _supported_manifest

    @ staticmethod
    @ Memoized
    def build_pipeline(yaml_file):
        pipelines = {}

        with open(yaml_file, 'r') as stream:
            try:
                pipelines_defs = yaml.safe_load(stream)

                # will raise exception if overall yml file config is invalid
                Config.validate(pipelines_defs)

                # A config file can contain a pipeline definition per manifest
                # format
                for pipeline_def in pipelines_defs:
                    manifest_format, plugin_def = next(
                        iter(pipeline_def.items())
                    )
                    plugins_list = []

                    for p in plugin_def:
                        plugin_name, plugin_config = next(iter(p.items()))
                        # Here you need to create the plugin
                        plugin = importlib.import_module(plugin_name)
                        importlib.invalidate_caches()

                        plugin_class_instance = getattr(plugin, "Plugin")()
                        plugin_class_instance.onCreate(
                            manifest_format, plugin_config)
                        if plugin_class_instance.validate_config(manifest_format):
                            plugins_list.append(plugin_class_instance)
                        else:
                            raise InvalidConfig('Invalid config section for plugin {}'.format(plugin_name))

                    pipelines[manifest_format] = Pipeline(
                        manifest_format, plugins_list)

            except yaml.YAMLError:
                raise Exception('Invalid Yaml file!')

            except ImportError:
                raise Exception(
                    'Unable to find python plugin defined in {}'.
                        format(yaml_file))

        # The yml configuration file can contain pipeline definitions just for
        #  one or few manifest format. Make sure that for each supported
        #  manifest format, there is at least an empty pipeline prepared.

        for format in Pipeline.SUPPORTED_MANIFEST_FORMATS:
            if not pipelines.get(format, None):
                pipelines[format] = []

        return pipelines

    def __init__(self, manifest_format, plugins_list):
        self.format = manifest_format
        self._pipeline = plugins_list

    def run(self, input_manifest=None, **kwargs):
        # Now process across the rest of the pipeline
        # The intent of having manifest_in as a read-only copy of the original
        # unmodified manifest is for the moment non possible until we implement
        # deepcopy (see above).
        for plugin in self._pipeline:
            plugin.process(manifest=input_manifest)
