from manifest_edit import BasePlugin
from manifest_edit import ManifestIterator
from ..context import Context
import re
from schema import Schema, Or


class Plugin(BasePlugin):
    '''
    manifest_order plugin.

    The purpose of this plugin is to rearrange Periods, Adaptation Sets or
    Representations in manifest files based on several possible criteria.

    Limitations: only implemented for .mpd manifests

    Uses ManifestIterator to "select" the parts that need to be reordered.

    Supported ordering strategies are:

    - full sort based on a provided key and "desc" or "asc" criteria
    - manual sorting, where the user must provide an ordered list of
      elements that need to match a given <key>, <regular expression> pair.
    '''

    _schema_full_sort = Schema(
        {
            'orderBy': Or(str, int),
            'orderCriteria': Or("desc", "asc")
        }
    )

    _schema_manual_order = Schema(
        {
            'manualOrder':
                [
                    {str: lambda regex: re.compile(regex)}
                ]
        }
    )

    _schema = Schema(Or(_schema_full_sort, _schema_manual_order))

    def validate_config(self, manifest_format):
        self.manifest_iterator = ManifestIterator(
            self.config, plugin_config_schema=self._schema)
        return self.manifest_iterator.is_valid()

    def reorderManifestLists(self, manifest):

        # We will just use manifest_iterator with all possible keywords.It will
        # return nothing in case the keyword is not present in self.config .
        # Maybe this can be optimized and only use it on what we know it's
        # present.
        for order_config, list_to_order in self.manifest_iterator(
                    manifest,
                    ['orderBy', 'orderCriteria']
                ):
            list_to_order.sort(
                key=lambda repr: getattr(
                    repr, order_config['orderBy'], ''),
                reverse=True if order_config['orderCriteria'] == 'desc' else False
            )

        for order_config, list_to_order in self.manifest_iterator(
                    manifest,
                    ['manualOrder']
                ):

            # iterate on list of key,value and take note of index of list
            for idx, config_item in enumerate(order_config['manualOrder']):
                # get from config the key to look for
                # get from config the re to match the value
                for key, expression in config_item.items():
                    # See if an element of the manifest matches the expression
                    # and, if yes, get its index in the list it belongs
                    regex = re.compile(expression)

                    found_at = [i for i, x in enumerate(list_to_order) if regex.search(
                        str(getattr(x, key, None)))]

                    if len(found_at):
                        # if a matching element is found, we will move it to
                        # the desired location. If it is found multiple times,
                        # only the first occurence is considered. If it is not
                        # found, nothing happens
                        list_to_order.insert(
                            idx, list_to_order.pop(found_at[0]))

                    # We now just break out of the config_item.items iteration
                    # because we don't support, for the moment, matching items
                    # by multiple key,expression couples.

    def mpd(self, manifest):
        self.reorderManifestLists(manifest)
